﻿namespace SystemHelpers
{
    using System;

    /// <summary>Postponable action.</summary>
    /// <copyright>Copyright © 2010 Andrey Veselov.</copyright>
    /// <example>Usage example:
    /// public class DemoSort {
    ///     public DemoSort() { this._postponableSort = this.SortImplementation; }
    ///     private PostponableAction2 _postponableSort;
    ///     private void SortImplementation()  { /* Implementation code here */ }
    ///     public void Sort() { this._postponableSort.Execute(); }
    ///     public IDisposable PostponeSort  { get { return this._postponableSort.PostponeAction; } }
    ///     
    ///     // Direct call example
    ///     public AddItem(object item) {
    ///         /* do something useful /*
    ///         this.Sort ();
    ///     }
    ///
    ///     // Using 'using' keyword
    ///     public PostponeDemoMethod() {
    ///         using (this.PostponeSort) {
    ///             this.AddItem (new object());
    ///             this.AddItem (new object());             
    ///         }
    ///     }
    /// }
    /// </example>
    public class PostponableAction
    {
        #region Private fields

        /// <summary>IDisposable interface helper object.</summary>
        private readonly DisposableInterfaceHelper _disposableInterfaceHelper;

        /// <summary>The action to be postponed.</summary>
        private Action _postponeAction;

        /// <summary>Postpone level counter.</summary>
        private int _postponeLevel = 0;

        /// <summary>Controls postponed action execution.</summary>
        private bool _mustExecute = false;

        /// <summary>Lock object (for thread safety).</summary>
        private object _lockObject = new object();

        #endregion

        /// <summary>Initializes a new instance of the PostponableAction class.</summary>
        /// <param name="sourceAction">The action to be postponed.</param>
        public PostponableAction(Action sourceAction)
        {
            if (sourceAction == null) {
                throw new ArgumentNullException("sourceAction");
            }

            this._disposableInterfaceHelper = new DisposableInterfaceHelper(this.OnDisposeHelper);
            this._postponeAction = sourceAction;
        }

#if DEBUG
        /// <summary>Finalizes an instance of the PostponableAction class. For debug only.</summary>
        ~PostponableAction()
        {
            if (this._postponeLevel != 0) {
                throw new InvalidOperationException("Postponed action planned but not executed.");
            }
        }
#endif

        #region Properties

        /// <summary>Gets a value indicating whether postponing is active.</summary>
        public bool IsPostponing
        {
            get { return this._postponeLevel > 0; }
        }

        #endregion

        /// <summary>Returns IDisposable interface. 
        /// This allow using the 'using' keyword to define where action will be postponed.</summary>
        /// <returns>IDisposable interface.</returns>
        public IDisposable PostponeAction()
        {
            lock (this._lockObject) {
                this._postponeLevel++;
            }

            return this._disposableInterfaceHelper;
        }

        /// <summary>Initiates the action execution. 
        /// If postponing is not active then action will be executed immediately.</summary>
        public void Execute()
        {
            lock (this._lockObject) {
                if (this._postponeLevel > 0) {
                    this._mustExecute = true;
                    return;
                }
            }

            this._postponeAction();
        }

        /// <summary>Executes postponed action if required.
        /// This method will be called from DisposableInterfaceHelper on dispose.</summary>
        private void OnDisposeHelper()
        {
            lock (this._lockObject) {
                this._postponeLevel--;
                if ((this._postponeLevel > 0) || !this._mustExecute) {
                    return;
                }

                this._mustExecute = false;
            }

            this._postponeAction();
        }

        /// <summary>Nested class. Perform an action when the instance is disposed.</summary>
        private class DisposableInterfaceHelper : IDisposable
        {
            /// <summary>Postponed action helper.</summary>
            private Action _onDispose;

            /// <summary>Initializes a new instance of the DisposableInterfaceHelper class.</summary>
            /// <param name="onDispose">The action to be called on dispose.</param>
            public DisposableInterfaceHelper(Action onDispose)
            {
                if (onDispose == null) {
                    throw new ArgumentNullException("onDispose");
                }

                this._onDispose = onDispose;
            }

            /// <summary>IDisposable implementation: Execute action on dispose.</summary>
            void IDisposable.Dispose()
            {
                this._onDispose();
            }
        }
    }
}
